from distutils.dir_util import copy_tree
import shutil
import glob
import os

def init_input_folder(self) -> bool:
    """
    Creating the tree structure for temporary files storage.
    By default it will be located in /tmp/frelatage
    """
    start_input_directory = self.input_directory
    tmp_input_directory = self.config.FRELATAGE_INPUT_FILE_TMP_DIR

    # Create /tmp/frelatage (default value) directory if not exists
    if not os.path.exists(tmp_input_directory):
        os.makedirs(tmp_input_directory)

    # Empty the directory
    else:
        list_dir = os.listdir(tmp_input_directory)
        for filename in list_dir:
            file_path = os.path.join(tmp_input_directory, filename)
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)

    # Create a subdirectory /tmp/frelatage/<subfolder> for each thread
    for thread in range(self.threads_count):
        thread_directory = os.path.join(tmp_input_directory, str(thread))
        os.makedirs(thread_directory)
        # Copy the input folder content into the subdirectory
        copy_tree(start_input_directory, thread_directory)
    return True

def init_file_input_arguments(self) -> bool:
    """
    Initialization of the file arguments value.

    Input {
        type: File
        value: <input_file_tmp_dir>/<thread 0>/<filename>
    }
    """
    file_input_arguments_count = 0
    # We initialize the value of each argument of type "file".
    for argument in self.arguments:
        if argument.file:
            # <input_file_tmp_dir>/<thread 0>/<filename>
            argument.value = "{base_directory}/{thread}/{file_input_arguments_count}".format(
                base_directory=self.config.FRELATAGE_INPUT_FILE_TMP_DIR,
                thread="0",
                file_input_arguments_count=file_input_arguments_count, 
            )
            file_input_arguments_count+= 1
    return True

def init_file_inputs(self) -> bool:
    """
    Set up the tree structure to fuzz a function with "file" type arguments
    """
    if all([argument.file for argument in self.arguments]):
        self.init_input_folder()
        self.init_file_input_arguments()


from frelatage.dictionnary.dictionnary import Dictionnary
import glob
import os

def load_dictionnary(dictionnary_folder: str) -> bool:
    """
    Load all the dictionnaries from the dictionnary folder
    """
    dictionnary = Dictionnary()
    dictionnaries = []
    if os.path.isdir(dictionnary_folder):
        # find all ".dict" files in the dictionnary folder
        dictionnary_files = glob.glob(os.path.join(dictionnary_folder, "*.dict"))
        for file in dictionnary_files:
            # parse dictionnary files
            dictionnaries += dictionnary.load_dictionnary_from_file(file)
    return dictionnaries


from threading import Thread

def fuzz(self) -> None:
    """
    Run the fuzzer
    """
    # Interface
    p = Thread(target=self.start_interface)
    p.start()

    # Fuzzing
    parents = [self.arguments]
    while True: 
        self.generate_cycle_mutations(parents)
        reports = self.run_cycle()
        parents = self.evaluate_mutations(reports)


from frelatage.report.report import Report
import pickle
import os

def get_report_name(self, report: Report) -> str:
    """
    Generate a report title.
    The name of a report is in the following form:
    id<crash ID>,err<error type>
    """
    # It is assumed that the number of uniques crashes will not exceed 999999
    error_id = str(self.unique_crashes).zfill(6) 
    error_type = report.trace.error_type.lower()

    report_name = "id:{error_id},err:{error_type}".format(
        error_id=error_id,
        error_type=error_type
    )
    return report_name

def save_report(self, report) -> bool:
    """
    Save a report in the output directory (/out by default).
    The arguments passed to the function and the behavior of the function are stored
    in "input", the file inputs are stored in files ranging from 0 to n.
    The report directory is in the following form: 
    ├── out
    │   ├── id<crash ID>,err<error type>
    │       ├── input
    │       ├── 0
    │       └── ...
    │   ├── ...
    """

    # The report contains the parameters passed to the function.
    custom_report = {
        "input": [dict(input) for input in report.input]
    }

    # /out by default
    base_directory = self.output_directory
    report_name = self.get_report_name(report)

    report_directory = "{base_directory}/{report_name}".format(base_directory=base_directory, report_name=report_name)

    # Create ./out/<report directory>/ directory if not exists
    if not os.path.exists(report_directory):
        os.makedirs(report_directory)
    
    # Save report file
    with open('{report_directory}/input'.format(report_directory=report_directory), 'wb+') as f:
        # We use pickle to store the report object
        pickle.dump(custom_report, f)

    # Save input files
    for argument in report.input:
        file_arguments_count = 0
        if argument.file:
            file_argument_content = open(argument.value, "rb").read()
            # Save file in /out/<report name>/<file id>
            with open("{report_directory}/{file_arguments_count}".format(report_directory=report_directory, file_arguments_count=str(file_arguments_count)), "wb+") as f:
                f.write(file_argument_content)
    return True


from dataclasses import dataclass

@dataclass
class Result:
    """
    Result of a trace
    """
    # True if an error occurred during the execution of a function, false otherwise
    error: bool
    # Error type, e.g TypeError, ZeroDivisionError...
    error_type: str
    # True if an TimeOutError occurred during the execution of a function, false otherwise
    timeout: bool
    # Position of the instruction where the error occured
    error_position: int
    # All groups of two successive instructions reached during the execution of a function
    # e.g : [(1, 2), (2, 5), (5, 6), (6, 3)]
    instructions_pairs: list
    # List of the reached instructions during the execution of a function
    reached_instructions: list[tuple]


import sys
import os 

class Config:
    """
    We can pass configuration through environnement variables
    """
    # Delay after which a function will return a timeoutError
    FRELATAGE_TIMEOUT_DELAY = int(os.getenv('TIMEOUT_DELAY', 2))
    # Temporary folder where input files are stored. 
    # Need to be an absolute path
    FRELATAGE_INPUT_FILE_TMP_DIR = str(os.getenv('FRELATAGE_INPUT_FILE_TMP_DIR', '/tmp/frelatage'))
    # Maximum size of an input variable
    FRELATAGE_INPUT_MAX_LEN = int(os.getenv('FRELATAGE_INPUT_MAX_LEN', 4096))
    # Maximum number of simultaneous threads
    FRELATAGE_MAX_THREADS = int(os.getenv('FRELATAGE_MAX_THREADS', 20))
    # Default directory for dictionaries
    # Relative path (to the path of the fuzzing file)
    FRELATAGE_DICTIONNARY_DIR = str(os.path.join(
                                        os.path.dirname(os.path.realpath(sys.argv[0])),
                                        os.getenv('FRELATAGE_DICTIONNARY_DIR', "./dict")
                                ))